(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * FTP.thy ---
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: FTP.thy 9221 2011-08-17 13:52:01Z lukasbru $ *)

theory FTP
imports 
  Stateful 
begin

text{* The File Transfer Protocol FTP is a well known example of a
  protocol which uses dynamic ports and is therefore a natural choice
  to use as an example for our model. 

  We model only a simplified version of the FTP protocol over IntegerPort
  addresses, still containing all messages that matter for our
  purposes. It consists of the following four messages:

  \begin{enumerate} 
    \item @{text "ftp_init"}: The client contacts the server indicating
          his wish to get some data.
    \item @{text "ftp_port_request p"}: The client, usually after having
          received an acknowledgement of the server, indicates a port
          number on which he wants to receive the data.
    \item @{text "ftp_data"}: The server sends the requested data over
          the new channel. There might be an arbitrary number of such
          messages, including zero.
      \item @{text "ftp_close"}: The client closes the connection. The
          dynamic port gets closed again.
  \end{enumerate}

  The content field of a packet therefore now consists of either one
  of those four messages or a default one.  *}


datatype ftp_msg = ftp_init 
                 | ftp_port_request port 
                 | ftp_data 
                 | ftp_close 
                 | other


text{* We now also make use of the ID field of a packet. It is used as
  session ID and we make the assumption that they are all unique among
  different protocol runs.

  At first, we need some predicates which check if a packet is a
  specific FTP message and has the correct session ID. 
*}

definition
  is_init :: "id \<Rightarrow> (IntegerPort, ftp_msg ) packet \<Rightarrow> bool" where 
  "is_init i p = (id p = i \<and> content p = ftp_init)"

definition
  is_port_request :: "id  \<Rightarrow> port \<Rightarrow>(IntegerPort, ftp_msg) packet \<Rightarrow> bool" where
  "is_port_request i port p = (id p = i \<and> content p = ftp_port_request port)" 

definition
  is_data :: "id \<Rightarrow> (IntegerPort, ftp_msg) packet \<Rightarrow> bool" where
  "is_data i p = (id p = i \<and> content p = ftp_data)"

definition
  is_close :: "id \<Rightarrow> (IntegerPort, ftp_msg) packet \<Rightarrow> bool" where
  "is_close i p = (id p = i \<and> content p = ftp_close)"

definition 
  port_open :: "(IntegerPort, ftp_msg) history \<Rightarrow> id \<Rightarrow> port \<Rightarrow> bool" where
  "port_open L a p = (not_before (is_close a) (is_port_request a p) L)"

definition
  is_other :: "id \<Rightarrow> (IntegerPort, ftp_msg ) packet \<Rightarrow> bool" where 
  "is_other i p = (id p = i \<and> content p = ftp_msg.other)"

fun are_other where
  "are_other i (x#xs) = (is_other i x \<and> are_other i xs)"
 |"are_other i [] = True"

text{*
  We now have to model the respective state transitions. It is
  important to note that state transitions themselves allow all
  packets which are allowed by the policy, not only those which are
  allowed by the protocol. Their only task is to change the policy. As
  an alternative, we could have decided that they only allow packets
  which follow the protocol (e.g. come on the correct ports), but this
  should in our view rather be reflected in the policy itself.

  Of course, not every message changes the policy. In such cases, we
  do not have to model different cases, one is enough. In our example,
  only messages 2 and 4 need special transitions. The default says
  that if the policy accepts the packet, it is added to the history,
  otherwise it is simply dropped. The policy remains the same in both
  cases.  *}


fun FTP_ST ::
  "((IntegerPort,ftp_msg) history, IntegerPort, ftp_msg) FWStateTransition"
where 
(* FTP_PORT_REQUEST *)
 "FTP_ST (i,s,d,ftp_port_request pr) (InL, policy) =
  (if p_accept (i,s,d,ftp_port_request pr) policy then
       (if not_before  (is_close i) (is_init i) InL \<and>
           dest_port (i,s,d,ftp_port_request pr) = (21::port) then
                          Some (allow (),((i,s,d,ftp_port_request pr)#InL, policy ++ 
                             (allow_from_to_port pr (subnet_of d) (subnet_of s))))
       else Some (allow (),((i,s,d,ftp_port_request pr)#InL,policy)))
  else Some (deny (),(InL,policy)))"

 
  |"FTP_ST (i,s,d,ftp_close) (InL,policy) = 
  (if (p_accept (i,s,d,ftp_close) policy) then 
       (if (\<exists> p. port_open InL i p)  \<and> dest_port (i,s,d,ftp_close) = (21::port) then 
             Some(allow (),((i,s,d,ftp_close)#InL, policy ++ 
     deny_from_to_port (Eps (\<lambda> p. port_open InL i p)) (subnet_of d) (subnet_of s)))
       else Some (allow (),((i,s,d,ftp_close)#InL, policy)))
  else Some (deny (),(InL,policy)))"


|"FTP_ST (i,s,d,ftp_close) (InL,policy) = 
  (if (p_accept (i,s,d,ftp_close) policy) then 
       (if (\<exists> p. port_open InL i p)  \<and> dest_port (i,s,d,ftp_close) = (21::port) then 
             Some(allow (),((i,s,d,ftp_close)#InL, policy ++ 
     deny_from_to_port (14::port) (subnet_of d) (subnet_of s)))
       else Some (allow (),((i,s,d,ftp_close)#InL, policy)))
  else Some (deny (),(InL,policy)))"
(* DEFAULT *)
  |"FTP_ST p x =  (if p_accept p (snd x) then 
                                              Some (allow (),(p#(fst x),snd x)) 
                                           else 
                                              Some (deny (),(fst x,snd x)))"

 


text{* The second message of the protocol is the port request. If the packet is
  allowed by the policy, and iff there is an opened but not
  yet closed FTP-Session with the same session ID, we change the
  policy such that the requested port is opened. If the policy allows
  the packet but there is no open protocol run, we do allow the packet
  but do not open the requested port.  

  In the last message, we need to close a port which we do not know
  directly. It has only been specified in a preceding port\_request
  message. Therefore a predicate is needed which checks if there is an
  open protocol run with an opened port.  This transition is the
  trickiest one. We need to close the port wich has been opened but
  not yet closed by a packet with the same session ID. Here we use the
  assumption that they are supposed to be unique. This transition
  introduces some kind of inconsistency. If the port that was
  requested was already open to start with, it gets closed here. The
  tester should be aware of this fact.

  This transition has also some other consequences. The Hilbert
  epsilon operator @{text "Eps"}, also written as @{text "SOME"},
  returns an arbitrary object for which the following predicate is
  @{text "True"} and is undefined otherwise. We use it to get the
  number of the port which we want to close. With the if-condition it
  is assured that such a port exists, but we might have problems if
  there are several of them. However, due to our assumption that the
  session IDs are unique, there won't be a problem as long as we do
  not open several ports in one single protocol run. This should not
  occur by the definition of the protocol, but if it does, which might
  happen if we want to test illegal protocol runs, some proof work
  might be needed.
*}





text{* 
  Now we specify our test scenario in more detail. We could test:
  \begin{itemize}
    \item one correct FTP-Protocol run,
    \item several runs after another,
    \item several runs interleaved,
    \item an illegal protocol run, or
    \item several illegal protocol runs.
  \end{itemize}

  We only do the the simplest case here: one correct protocol run.
*}

text{*
  There are four different states which are modelled as a datatype.
*}

datatype ftp_states = S0 | S1 | S2 | S3


text{* 
  The following constant is @{text "True"} for all sets which are
  correct FTP runs for a given source and destination address, ID, and
  data-port number.  
*}



fun
  is_ftp :: "ftp_states \<Rightarrow> IntegerPort  \<Rightarrow> IntegerPort \<Rightarrow> id \<Rightarrow> port \<Rightarrow>
            (IntegerPort,ftp_msg) history \<Rightarrow> bool"
where
 "is_ftp H c s i p [] = (H=S3)"
|"is_ftp H c s i p (x#InL) = (\<lambda> (id,sr,de,co). (((id = i \<and> (
   (H=S2 \<and> sr = c \<and> de = s \<and> co = ftp_init \<and> is_ftp S3 c s i p InL) \<or> 
   (H=S1 \<and> sr = c \<and> de = s \<and> co = ftp_port_request p \<and>  is_ftp S2 c s i p InL) \<or> 
   (H=S1 \<and> sr = s \<and> de = (fst c,p) \<and> co= ftp_data \<and> is_ftp S1 c s i p InL) \<or> 
    (H=S0 \<and> sr = c \<and> de = s \<and> co = ftp_close  \<and> is_ftp S1 c s i p InL) ))))) x"
text{* 
  This definition is crucial for specifying what we actually want to
  test. Extending it produces more test cases but increases the time
  necessary to create them and vice-versa.  
*}




text{* 
  The following constant then returns a set of all the historys which
  denote such a normal behaviour FTP run, again for a given source and
  destination address, ID, and data-port. 
*}
definition 
  NB_ftp :: "IntegerPort src \<Rightarrow> IntegerPort dest \<Rightarrow> id \<Rightarrow> port  \<Rightarrow>
            (IntegerPort,ftp_msg) history set" where
  "NB_ftp s d i p = {x. (is_ftp S0 s d i p x)}"

text{* The following definition returns the set of all possible
interleaving of two correct FTP protocol runs. *}

definition 
  ftp_2_interleaved :: "IntegerPort src \<Rightarrow> IntegerPort dest \<Rightarrow> id \<Rightarrow> port  \<Rightarrow>
               IntegerPort src \<Rightarrow> IntegerPort dest \<Rightarrow> id \<Rightarrow> port  \<Rightarrow>
            (IntegerPort,ftp_msg) history set" where
  "ftp_2_interleaved s1 d1 i1 p1 s2 d2 i2 p2 = 
      {x. (is_ftp ftp_states.S0 s1 d1 i1 p1 (packet_with_id x i1)) \<and>
          (is_ftp ftp_states.S0 s2 d2 i2 p2 (packet_with_id x i2))}"




lemma subnetOf_lemma: "(a::int) \<noteq> (c::int) \<Longrightarrow> \<forall>x\<in>subnet_of (a, b::port). (c, d) \<notin> x"
apply (rule ballI)
apply (simp add: IntegerPort.subnet_of_def)
done

lemma subnetOf_lemma2: " \<forall>x\<in>subnet_of (a::int, b::port). (a, b) \<in>  x"
apply (rule ballI)
apply (simp add: IntegerPort.subnet_of_def)
done

lemma subnetOf_lemma3: "(\<exists>x. x \<in> subnet_of (a::int, b::port))"
apply (rule exI)
apply (simp add: IntegerPort.subnet_of_def)
done

lemma subnetOf_lemma4: "\<exists>x\<in>subnet_of (a::int, b::port). (a, c::port) \<in> x"
apply (rule bexI)
apply (simp_all add: IntegerPort.subnet_of_def)
done

lemma port_open_lemma: "\<not> (Ex (port_open [] (x::port)))"
apply (simp add: port_open_def)
done

end
